"use strict";
const tslib_1 = require("tslib");
const util_1 = tslib_1.__importDefault(require("util"));
const fs = tslib_1.__importStar(require("fs-extra"));
const path_1 = require("path");
const lisk_db_1 = require("@liskhq/lisk-db");
const semver = tslib_1.__importStar(require("semver"));
const command_1 = require("@oclif/command");
const cli_ux_1 = tslib_1.__importDefault(require("cli-ux"));
const constants_1 = require("./constants");
const client_1 = require("./client");
const config_1 = require("./utils/config");
const chain_1 = require("./utils/chain");
const events_1 = require("./events");
const genesis_block_1 = require("./utils/genesis_block");
const createAsset_1 = require("./createAsset");
const node_1 = require("./utils/node");
const fs_1 = require("./utils/fs");
const process_1 = require("./utils/process");
let configCoreV4;
class LiskMigrator extends command_1.Command {
    async run() {
        var _a, _b, _c;
        try {
            const { flags } = this.parse(LiskMigrator);
            const liskCoreV3DataPath = (0, fs_1.resolveAbsolutePath)((_a = flags['lisk-core-v3-data-path']) !== null && _a !== void 0 ? _a : constants_1.DEFAULT_LISK_CORE_PATH);
            const outputPath = (_b = flags.output) !== null && _b !== void 0 ? _b : (0, path_1.join)(__dirname, '..', 'output');
            const snapshotHeight = flags['snapshot-height'];
            const customConfigPath = flags.config;
            const autoMigrateUserConfig = (_c = flags['auto-migrate-config']) !== null && _c !== void 0 ? _c : false;
            const autoStartLiskCoreV4 = flags['auto-start-lisk-core-v4'];
            const pageSize = Number(flags['page-size']);
            const client = await (0, client_1.getAPIClient)(liskCoreV3DataPath);
            const nodeInfo = (await client.node.getNodeInfo());
            const { version: appVersion, networkIdentifier } = nodeInfo;
            cli_ux_1.default.action.start('Verifying if backup height from node config matches snapshot height');
            if (snapshotHeight !== nodeInfo.backup.height) {
                this.error(`Lisk Core v3 backup height (${nodeInfo.backup.height}) does not match the expected snapshot height (${snapshotHeight}).`);
            }
            cli_ux_1.default.action.stop('Snapshot height matches backup height');
            cli_ux_1.default.action.start(`Verifying snapshot height to be multiples of round length i.e ${constants_1.ROUND_LENGTH}`);
            if (snapshotHeight % constants_1.ROUND_LENGTH !== 0) {
                this.error(`Invalid Snapshot Height: ${snapshotHeight}.`);
            }
            cli_ux_1.default.action.stop('Snapshot height is valid');
            const networkConstant = constants_1.NETWORK_CONSTANT[networkIdentifier];
            const outputDir = flags.output ? outputPath : `${outputPath}/${networkIdentifier}`;
            if (!fs.existsSync(outputDir))
                fs.mkdirSync(outputDir, { recursive: true });
            (0, events_1.captureForgingStatusAtSnapshotHeight)(this, client, snapshotHeight, outputDir);
            if (autoStartLiskCoreV4) {
                if (!networkConstant) {
                    this.error(`Unknown network detected. No NETWORK_CONSTANT defined for networkID: ${networkIdentifier}.`);
                }
            }
            cli_ux_1.default.action.start('Verifying Lisk Core version');
            const isLiskCoreVersionValid = semver.valid(appVersion);
            if (isLiskCoreVersionValid === null) {
                this.error(`Invalid Lisk Core version detected: ${appVersion}. Minimum supported version is ${constants_1.MIN_SUPPORTED_LISK_CORE_VERSION}.`);
            }
            if (semver.gt(`${constants_1.MIN_SUPPORTED_LISK_CORE_VERSION}-rc.0`, appVersion)) {
                this.error(`Lisk Migrator is not compatible with Lisk Core version ${appVersion}. Minimum supported version is ${constants_1.MIN_SUPPORTED_LISK_CORE_VERSION}.`);
            }
            cli_ux_1.default.action.stop(`${appVersion} detected`);
            const configV3 = customConfigPath
                ? await (0, config_1.getConfig)(liskCoreV3DataPath, customConfigPath)
                : await (0, config_1.getConfig)(liskCoreV3DataPath);
            await (0, chain_1.setTokenIDLskByNetID)(networkIdentifier);
            await (0, chain_1.setPrevSnapshotBlockHeightByNetID)(networkIdentifier);
            await (0, chain_1.observeChainHeight)({
                label: 'Waiting for snapshot height to be finalized',
                liskCoreV3DataPath,
                height: snapshotHeight,
                delay: 500,
                isFinal: true,
            });
            cli_ux_1.default.action.start('Creating database instance');
            const snapshotDirPath = (0, path_1.join)(liskCoreV3DataPath, constants_1.SNAPSHOT_DIR);
            const db = new lisk_db_1.Database(snapshotDirPath);
            cli_ux_1.default.action.stop();
            cli_ux_1.default.action.start('Creating genesis assets');
            const createAsset = new createAsset_1.CreateAsset(db);
            const tokenID = (0, chain_1.getTokenIDLsk)();
            const genesisAssets = await createAsset.init(snapshotHeight, tokenID, pageSize);
            cli_ux_1.default.action.stop();
            const defaultConfigFilePath = await (0, config_1.resolveConfigPathByNetworkID)(networkIdentifier);
            const defaultConfigV4 = await fs.readJSON(defaultConfigFilePath);
            cli_ux_1.default.action.start(`Exporting genesis block to the path ${outputDir}`);
            await (0, genesis_block_1.writeGenesisAssets)(genesisAssets, outputDir);
            cli_ux_1.default.action.stop();
            if (autoMigrateUserConfig) {
                cli_ux_1.default.action.start('Creating backup for old config');
                await (0, config_1.createBackup)(configV3);
                cli_ux_1.default.action.stop();
                cli_ux_1.default.action.start('Migrating user configuration');
                const migratedConfigV4 = (await (0, config_1.migrateUserConfig)(configV3, defaultConfigV4, snapshotHeight));
                cli_ux_1.default.action.stop();
                cli_ux_1.default.action.start('Validating migrated user configuration');
                const isValidConfig = await (0, config_1.validateConfig)(migratedConfigV4);
                cli_ux_1.default.action.stop();
                if (!isValidConfig)
                    throw new Error('Migrated user configuration is invalid.');
                cli_ux_1.default.action.start(`Exporting user configuration to the path: ${outputDir}`);
                await (0, config_1.writeConfig)(migratedConfigV4, outputDir);
                cli_ux_1.default.action.stop();
                configCoreV4 = migratedConfigV4;
            }
            cli_ux_1.default.action.start('Installing Lisk Core v4');
            await (0, node_1.installLiskCore)();
            cli_ux_1.default.action.stop();
            cli_ux_1.default.action.start('Creating genesis block');
            const blockAtSnapshotHeight = (await client.block.getByHeight(snapshotHeight));
            await (0, genesis_block_1.createGenesisBlock)(networkConstant.name, defaultConfigFilePath, outputDir, blockAtSnapshotHeight);
            cli_ux_1.default.action.stop();
            if (autoStartLiskCoreV4) {
                try {
                    if (!autoMigrateUserConfig) {
                        configCoreV4 = defaultConfigV4;
                    }
                    cli_ux_1.default.action.start('Copying genesis block to the Lisk Core executable directory');
                    const liskCoreExecPath = await (0, process_1.execAsync)('which lisk-core');
                    const liskCoreV4ConfigPath = (0, path_1.resolve)(liskCoreExecPath, '../..', `lib/node_modules/lisk-core/config/${networkConstant.name}`);
                    await (0, genesis_block_1.copyGenesisBlock)(`${outputDir}/genesis_block.blob`, `${liskCoreV4ConfigPath}/genesis_block.blob`);
                    this.log(`Genesis block has been copied to: ${liskCoreV4ConfigPath}.`);
                    cli_ux_1.default.action.stop();
                    const isLiskCoreV3Stopped = await cli_ux_1.default.confirm("Please stop Lisk Core v3 to continue. Type 'yes' and press Enter when ready. [yes/no]");
                    if (isLiskCoreV3Stopped) {
                        const isUserConfirmed = await cli_ux_1.default.confirm(`Start Lisk Core with the following configuration? [yes/no] \n${util_1.default.inspect(configCoreV4, false, 3)}`);
                        if (isUserConfirmed) {
                            cli_ux_1.default.action.start('Starting Lisk Core v4');
                            const network = networkConstant.name;
                            await (0, node_1.startLiskCore)(this, liskCoreV3DataPath, configCoreV4, network, outputDir);
                            this.log('Started Lisk Core v4 at default data directory.');
                            cli_ux_1.default.action.stop();
                        }
                        else {
                            this.log('User did not accept the migrated config. Skipping the Lisk Core v4 auto-start process.');
                        }
                    }
                    else {
                        this.log('User did not confirm Lisk Core v3 node shutdown. Skipping the Lisk Core v4 auto-start process.');
                    }
                }
                catch (err) {
                    this.error(`Failed to auto-start Lisk Core v4.\nError: ${err}`);
                }
            }
            else {
                this.log(`Please copy the contents of ${snapshotDirPath} directory to 'data/legacy.db' under the Lisk Core V4 data directory (e.g: ~/.lisk/lisk-core/data/legacy.db/) in order to access legacy blockchain information.`);
                this.log('Please copy genesis block to the Lisk Core V4 network directory.');
            }
        }
        catch (error) {
            this.error(error);
        }
        this.log('Successfully finished migration. Exiting!!!');
        process.exit(0);
    }
}
LiskMigrator.description = 'Migrate Lisk Core to latest version';
LiskMigrator.flags = {
    version: command_1.flags.version({ char: 'v' }),
    help: command_1.flags.help({ char: 'h' }),
    output: command_1.flags.string({
        char: 'o',
        required: false,
        description: "File path to write the genesis block. If not provided, it will default to cwd/output/{v3_networkIdentifier}/genesis_block.blob. Do not use any value starting with the default data path reserved for Lisk Core: '~/.lisk/lisk-core'.",
    }),
    'lisk-core-v3-data-path': command_1.flags.string({
        char: 'd',
        required: false,
        description: 'Path where the Lisk Core v3.x instance is running. When not supplied, defaults to the default data directory for Lisk Core.',
    }),
    config: command_1.flags.string({
        char: 'c',
        required: false,
        description: 'Custom configuration file path for Lisk Core v3.x.',
    }),
    'snapshot-height': command_1.flags.integer({
        char: 's',
        required: true,
        env: 'SNAPSHOT_HEIGHT',
        description: 'The height at which re-genesis block will be generated. Can be specified with SNAPSHOT_HEIGHT as well.',
    }),
    'auto-migrate-config': command_1.flags.boolean({
        required: false,
        env: 'AUTO_MIGRATE_CONFIG',
        description: 'Migrate user configuration automatically. Default to false.',
        default: false,
    }),
    'auto-start-lisk-core-v4': command_1.flags.boolean({
        required: false,
        env: 'AUTO_START_LISK_CORE',
        description: 'Start lisk core v4 automatically. Default to false.',
        default: false,
    }),
    'page-size': command_1.flags.integer({
        char: 'p',
        required: false,
        default: 100000,
        description: 'Maximum number of blocks to be iterated at once for computation. Default to 100000.',
    }),
};
module.exports = LiskMigrator;
//# sourceMappingURL=index.js.map